// Copyright 2003 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "config.h"
#undef USE_MYSQL
#define loc_t sys_loc_t
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <algorithm>
#include <vector>
#undef loc_t

#include "sqlint8.h"
#include "sqlca.h"
#include "sqlda.h"
#include "sqlhdr.h"
#include "locator.h"
#include "sqltypes.h"
#include "sqlstype.h"
#include "sqldefs.h"
#include "sqlint8.h"
#include "sqlblob.h"
#include "sqlrow.h"
#include "sqlapi.h"

extern sqlca_s sqlca;

#ifdef USE_NAMESPACE
namespace ifx {
#endif
typedef struct {
  int type;
  char *byte_ptr;
}
BLOBENV;

typedef struct {
  char       *cmd;
unsigned int opened:1;
unsigned int is_select:1;
  struct sqlda *in;
  struct sqlda *out;
  struct sqlda *exec;
  int *arr_sqltype;
  int arr_sqltype_len;
  int *arr_sqllen;
  char          p_name[19];
  char          c_name[19];
  sqlint8_t insert_id;
} SSQL;

static char *null_cmd="";

#define debug(s)    s

#define errmalloc(pointer, func, len)   \
    if (!pointer) {\
        sprintf(errmsg, "Cannot malloc for %uld in %s errno %d",len,func,errno);\
        debug(fprintf(stderr, "%s\n", errmsg));\
        return false;\
    }

#define make_into_char(arg) \
    (((arg) >= 'A' && (arg) <= 'F')?((arg) - 'A' + 10):\
    (((arg) >= 'a' && (arg) <= 'f')?((arg) - 'a' + 10):(arg)))


#define chk_status(p1, p2)  if (sqlca.sqlcode != 0) {\
    if (sqlca.sqlcode<0) { \
       sql_close(fd); \
       return print_status(p1, p2); \
    } \
    print_status(p1,p2); \
}


static SSQL ssql[sql::MAXCURSORS+1];
static char *database_name;
static char *connection_name;

static char errmsg[1024+1];

#ifndef lint
static char const rcs[] = "@(#)$Id: sqlifx.ec,v 1.19 2004/11/03 00:21:01 korpela Exp $";
#endif

sqlint8_t sql_insert_id() {
  sqlint8_t rv;
  ifx_int8_t val;
  char a[21];
  ifx_getserial8(&val);
  ifx_int8toasc(&val,&a[0],24);
  a[20]=0;
  sscanf(a,"%"INT8_FMT,&rv);
  if (!rv) {
    rv=static_cast<sqlint8_t>(sqlca.sqlerrd[1]);
  }
  return rv;
}

#define INSERT_ID sql_insert_id()

sqlint8_t sql_insert_id(SQL_CURSOR fd) {
  return ssql[fd].insert_id;
}


bool chk_cursor(SQL_CURSOR fd, const char *function) {
  if (fd < 0 || fd >= sql::MAXCURSORS || !ssql[fd].cmd) {
    sprintf(errmsg, "Invalid fd %d passed to %s()", fd, function);
    debug(fprintf(stderr, "%s\n", errmsg));
    return false;
  }
  return true;
}

/* removes trailing white spaces */
static void
strip(char *s, int len) {
  int i;
  if (len > 0) {
    for (i = len - 1; i >=0 ; i--) {
      if (!isspace(s[i]))
        break;
    }
    s[i+1] = 0;
  }
}

/* Using the previous SQL error code, format the Informix error message */
static bool
print_status(const char *p1, const char *p2) {
  int         retcode = sqlca.sqlcode;
  if (retcode < 0) {
    int msglen;
    errmsg[0] = '\0';
    rgetlmsg(sqlca.sqlcode, errmsg, sizeof(errmsg), &msglen);
    if (sqlca.sqlerrd[1]) {
      int len = strlen(errmsg);
      errmsg[len] = '\0';
      rgetlmsg(sqlca.sqlerrd[1], &errmsg[len], sizeof(errmsg) - len, &msglen);
    }
    strip(sqlca.sqlerrp, strlen(sqlca.sqlerrp));
    strip(sqlca.sqlerrm, strlen(sqlca.sqlerrm));
    sprintf(&errmsg[strlen(errmsg)],
            "%ld:%ld:%s %s %s %s\n",
            sqlca.sqlcode, sqlca.sqlerrd[1], sqlca.sqlerrp, sqlca.sqlerrm, p1, p2);
    debug(fprintf(stderr, "%s\n", errmsg));
  }
  return (retcode>=0);
}

/* Set explain on or off */
bool sql_explain(int n) {
  int fd=0;
  if (n) {
    $set explain on;
    chk_status("sql_explain","set explain on");
  } else {
    $set explain off;
    chk_status("sql_explain","set explain off");
  }
  return (sqlca.sqlcode==0);
}

/* Return the error message stored in the buffer */
const char *sql_error_message(void) {
  return errmsg;
}

/*
** Check whether column (field) exists in table with value specified by val
** subject to additional constraints specified in where.
**
** returns
**  true if field exists in table,
**  false otherwise
*/
static int alloc_stmt();

bool sql_exists(const char *table, const char *field, const char *p_val, const char *where) {
  $char *value = const_cast<char *>(p_val);
  $char stmt[sql::BUFSIZ+1];
  $char result[2];
  int fetch_code = 0; /* NOTUSED */
  int fd=alloc_stmt();

  if (value && *value != '\0') {
    sprintf(stmt, "select '1' from %s where %s = ?",
            table, field);
    if (where && *where != '\0')
      sprintf(&stmt[strlen(stmt)], " and %s", where);
  } else {
    sprintf(stmt, "select '1' from %s",
            table);
    if (where && *where != '\0')
      sprintf(&stmt[strlen(stmt)], " where %s", where);
  }

  $prepare pexist from $stmt;
  chk_status("PREPARE(EXISTS)", stmt);
  $declare pexist_cur cursor for pexist;
  chk_status("DECLARE(EXISTS)", stmt);
  if (value && *value != '\0')
    $open pexist_cur using $value;
  else
    $open pexist_cur;
  chk_status("OPEN(EXISTS)", stmt);
  $fetch pexist_cur into $result;
  fetch_code = sqlca.sqlcode;
  $close pexist_cur;
  sql_close(fd);
  return (fetch_code==0);
}

static int convert_byte(char *tgt, const char *src, int len) {
  int i;
  int j;

  for (i = j = 0; i < len; i += 2) {
    tgt[j++] = make_into_char(src[i])*16 + make_into_char(src[i+1]);
  }
  tgt[j] = '\0';
  return j-1;
}

static bool sql_set_values(SQL_CURSOR fd, SQL_ROW &argv) {
  struct sqlvar_struct *col;
  struct sqlvar_struct *ecol;
  struct sqlda *udesc;
  struct sqlda *edesc;
  ifx_sqlvar_t *tmp=0;
  int i;
  int argc=argv.argc();

  udesc = ssql[fd].in;
  edesc = ssql[fd].exec;
  if (argc > 0) {
    if (edesc && (argc > edesc->sqld)) {
      tmp=edesc->sqlvar;
      edesc->sqlvar=(ifx_sqlvar_t *)calloc(argc, sizeof(ifx_sqlvar_t));
      memcpy(edesc->sqlvar,tmp,edesc->sqld*sizeof(ifx_sqlvar_t));
    }
    if (udesc && argc != udesc->sqld) {
      sprintf(errmsg,
              "Incorrect number (%d) of bind variables specified - %d required",
              argc, udesc->sqld);
      debug(fprintf(stderr, "%s\n", errmsg));
      if (tmp) {
        free(edesc->sqlvar);
        edesc->sqlvar=tmp;
      }
      return false;
    }
    if (udesc) {
      int i;
      for (i = 0, col = udesc->sqlvar; i < udesc->sqld; i++, col++) {
        if (col->sqltype == CLOCATORTYPE) {
          loc_t *loc = (loc_t *)col->sqldata;
          if (loc && loc->loc_buffer)
            free(loc->loc_buffer), loc->loc_buffer = NULL;
          if (loc && loc->loc_user_env) {
            BLOBENV *benv = (BLOBENV *)loc->loc_user_env;
            if (benv->byte_ptr) free(benv->byte_ptr);
            benv->byte_ptr = NULL;
            free(loc->loc_user_env), loc->loc_user_env = NULL;
          }
        }
        free(col->sqldata);
        col->sqldata = NULL;
      }
      free(udesc->sqlvar);
      free(udesc);
    }
    ssql[fd].in = (struct sqlda *)calloc(1, sizeof(struct sqlda));
    if (!ssql[fd].in) sql_close(fd);
    errmalloc(ssql[fd].in, "SETVALUES", sizeof(struct sqlda));
    memset(ssql[fd].in, 0, sizeof(struct sqlda));
    (ssql[fd].in)->sqld = argc;
    udesc = (ssql[fd].in);
    udesc->sqlvar = (struct sqlvar_struct *)
                    calloc(argc, sizeof(struct sqlvar_struct));
    if (!udesc->sqlvar)
      sql_close(fd);
    errmalloc(udesc->sqlvar, "SETVALUES",
              argc * sizeof(struct sqlvar_struct));
    memset(udesc->sqlvar, 0, argc * sizeof(struct sqlvar_struct));
    if (edesc && edesc->sqld <= 0) edesc = NULL;
    if (edesc && edesc->sqlvar) ecol = edesc->sqlvar;
    else ecol = NULL;
    for (i = 0, col = udesc->sqlvar; i < argc; i++, col++) {
      int blob = 0;
      size_t bloblen;
      if (ecol) {
        if ((ecol->sqltype&SQLTYPE) == SQLBYTES) blob = 1;
        if ((ecol->sqltype&SQLTYPE) == SQLTEXT) blob = 2;
        if (blob) col->sqldata = (char *)calloc(1, sizeof(loc_t));
      }
      if (argv[i]) {
      /* remove leading and trailing single quotes */
        char c,d;
        if (((c=argv[i][0])=='\'') && ((d=argv[i][strlen(argv[i])-1])=='\'')) {
	  argv[i][strlen(argv[i])-1]=0;
	  memmove(argv[i],argv[i]+1,strlen(argv[i]));
        }
      /* Allow <BYTE> and <TEXT> tags in values */
        if (strncmp(argv[i], "<BYTE", 5) == 0) {
          blob = 1;
	  long long etag=strchr(argv[i],'>')-argv[i]+1;
	  if (etag>0) {
	    char *p=strchr(argv[i],'=');
	    int tmp=strlen(argv[i]+etag);
	    if (p && (p<(argv[i]+etag))) sscanf(p+1,"%d",&tmp);
	    bloblen=tmp;
	    argv[i]+=etag;
	  } else {
	    argv[i]+=6;
	    bloblen=strlen(argv[i])+1;
	  }
          col->sqllen = sizeof(loc_t);
          col->sqldata = (char *)calloc(1, sizeof(loc_t));
        } else if (strncmp(argv[i], "<TEXT>", 6) == 0) {
          blob = 2;
          argv[i] = argv[i] + 6; /* Skip over the tag */
	  bloblen=strlen(argv[i])+1;
          col->sqllen = sizeof(loc_t);
          col->sqldata = (char *)calloc(1, sizeof(loc_t));
        }
      }
      if (blob) {
        loc_t *loc;
        BLOBENV *benv;
        if (!col->sqldata) sql_close(fd);
        errmalloc(col->sqldata, "SETVALUES LOC", sizeof(loc_t));
        loc = (loc_t *)col->sqldata;
        loc->loc_loctype = LOCMEMORY;
        loc->loc_buffer = argv[i];
        loc->loc_bufsize = loc->loc_size = bloblen;
        loc->loc_oflags = loc->loc_mflags = 0;
        col->sqltype = CLOCATORTYPE;
        benv = (BLOBENV *)calloc(1, sizeof(BLOBENV));
        if (!benv) sql_close(fd);
        errmalloc(benv, "SETVALUES BLOBENV", sizeof(BLOBENV));
        loc->loc_user_env = (char *)benv;
        if (blob == 1) {
          int byte_size;;
          benv->type = 'B';
          benv->byte_ptr =
            (char *)calloc(loc->loc_size, sizeof(char));
          if (!benv->byte_ptr) sql_close(fd);
          errmalloc(benv->byte_ptr, "SETVALUES BENV",
                    sizeof(BLOBENV));
          byte_size = convert_byte(benv->byte_ptr, loc->loc_buffer,
                                   loc->loc_size);
          loc->loc_buffer = benv->byte_ptr;
          loc->loc_bufsize = loc->loc_size = byte_size;
        } else {
          benv->type = 'T';
          loc->loc_buffer = (char *)calloc(loc->loc_bufsize,1);
          if (!loc->loc_buffer)
            sql_close(fd);
          errmalloc(loc->loc_buffer, "SETVALUES", (unsigned)loc->loc_bufsize);
          strcpy(loc->loc_buffer, argv[i]);
        }
      } else {
        col->sqllen = strlen(argv[i]);
        col->sqldata = (char *)calloc(col->sqllen+1, 1);
        if (!col->sqldata)
          sql_close(fd);
        errmalloc(col->sqldata, "SETVALUES",(unsigned)(col->sqllen+1));
        strcpy(col->sqldata, argv[i]);
        col->sqltype = CCHARTYPE;
      }
      col->sqlname = 0;
      col->sqlind = 0;
      if (ecol) ecol++;
    }
  }
  if (tmp) {
    free(edesc->sqlvar);
    edesc->sqlvar=tmp;
  }
  return true;
}

/* Returns
    true on successful execution
    sqlca.sqlcode otherwise
*/
bool sql_run(const char *stmt, SQL_ROW &argv) {
  SQL_CURSOR fd;
  int ret;

  if (isdigit(stmt[0])) {
    fd = atoi(stmt);
    if (!chk_cursor(fd, "sql_run"))
      sql_close(fd);
      return false;
    if (!sql_set_values(fd, argv))
      sql_close(fd);
      return false;
    return sql_fetch(fd);
  }
  fd = sql_open(stmt, argv);
  if (fd < 0) {
      sql_close(fd);
      return false;
  }
  ret = sql_fetch(fd);
  sql_close(fd);
  return ret;
}

/*
** Allocate a statement number
** Returns:
**   -1: no available statements
**   0..sql::MAXCURSORS-1: new statement number
*/
static int alloc_stmt() {
  int i;

  for (i = 0; i < sql::MAXCURSORS; i++) {
    if (!ssql[i].cmd)
      break;
  }
  if (i >= sql::MAXCURSORS) {
    sprintf(errmsg,
            "FATAL ERROR: Too many open SQL statements (> %d) for sql operations\n", sql::MAXCURSORS);
    debug(fprintf(stderr, "%s\n", errmsg));
    i = -1;
    exit(1);
  } else {
    ssql[i].cmd=null_cmd;
    sprintf(ssql[i].p_name, "p_%d", i);
    sprintf(ssql[i].c_name, "c_%d", i);
  }
  return(i);
}

SQL_CURSOR sql_open(const char *sql, SQL_ROW &argv) {
  struct sqlvar_struct *col = 0L;
  struct sqlda *udesc = 0L;
  $char *stmt = const_cast<char *>(sql);
  int         i;
  int         fd;
  int         len;
  char       *save;
  $char *p_name;
  $char *c_name;

  if ((fd = alloc_stmt()) < 0)
    return(-1);

  len = strlen(stmt);
  save = (char *) calloc(len + 1, 1);
  errmalloc(save, "OPEN", len);
  strcpy(save, stmt);
  ssql[fd].cmd = save;

  p_name = ssql[fd].p_name;

  $prepare $p_name from $stmt;

  /*
  if (sqlca.sqlcode < 0)
      sql_close(fd);
  */
  chk_status("PREPARE(OPEN)", stmt);

  $describe $p_name into udesc;

  chk_status("DESCRIBE(OPEN)", stmt);

  /**
  ** The cursory statements are:
  ** -- EXECUTE PROCEDURE
  ** -- SELECT without INTO TEMP
  ** Fortunately, SELECT INTO TEMP is given code SQ_SELINTO
  ** Unfortunately, SELECT is not given code SQ_SELECT (2) but 0
  ** Non-cursory statements need minimal extra attention.
  */
  if (sqlca.sqlcode != 0 &&
      (sqlca.sqlcode != SQ_EXECPROC || udesc->sqld == 0) &&
      sqlca.sqlcode != SQ_SELECT) {
    ssql[fd].is_select = 0;
    ssql[fd].exec = udesc;
    if (!sql_set_values(fd, argv))
      return sql_close(fd), -1;
    return fd;
  }

  ssql[fd].is_select = 1;
  ssql[fd].out = udesc;

  if (!udesc) {
    sprintf(errmsg, "No columns selected !\n");
    fprintf(stderr, "No columns selected !\n");
    return -1;
  }

  for (i = 0, col = udesc->sqlvar; i < udesc->sqld; i++, col++) {
    int         blob = 0;
    if ((col->sqltype & SQLTYPE) != SQLCHAR &&
        (col->sqltype & SQLTYPE) != SQLVCHAR) {
      if ((col->sqltype & SQLTYPE) == SQLBYTES) {
        col->sqldata = (char *) calloc(1, sizeof(loc_t));
        blob = 1;
      } else if ((col->sqltype & SQLTYPE) == SQLTEXT) {
        col->sqldata = (char *) calloc(1, sizeof(loc_t));
        blob = 2;
      } else if ((col->sqltype & SQLTYPE) == SQLROW) {
        col->sqllen = sql::MAXBLOB;
      } else if ((col->sqltype & SQLTYPE) == SQLLIST) {
        col->sqllen = sql::MAXBLOB;
      } else {
        col->sqllen = rtypwidth(col->sqltype, col->sqllen);
      }
      /* Should allocate at least 12 characters for backward compat. ! */
      if (col->sqllen < 12)
        col->sqllen = 12;
    }
    if (blob) {
      loc_t      *loc;
      BLOBENV    *benv;
      if (!col->sqldata)
        sql_close(fd);
      errmalloc(col->sqldata, "OPEN LOC", sizeof(loc_t));
      loc = (loc_t *) col->sqldata;
      loc->loc_loctype = LOCMEMORY;
      loc->loc_bufsize = -1;
      loc->loc_oflags = loc->loc_mflags = 0;
      col->sqltype = CLOCATORTYPE;
      benv = (BLOBENV *) calloc(1, sizeof(BLOBENV));
      if (!benv)
        sql_close(fd);
      errmalloc(benv, "OPEN BLOBENV", sizeof(BLOBENV));
      loc->loc_user_env = (char *) benv;
      if (blob == 1)
        benv->type = 'B';
      else
        benv->type = 'T';
    } else {
      col->sqllen++;
      col->sqldata = (char *) calloc(col->sqllen + 1, 1);
      if (!col->sqldata)
        sql_close(fd);
      errmalloc(col->sqldata, "OPEN", (int)col->sqllen + 1);
      memset(col->sqldata, 0, col->sqllen + 1);
      col->sqltype = CCHARTYPE;
    }
    col->sqlind = (short *) calloc(1, sizeof(short));
    if (!col->sqlind)
      sql_close(fd);
    errmalloc(col->sqlind, "OPEN", sizeof(short));
    *(col->sqlind) = 0;
  }

  c_name = ssql[fd].c_name;

  $declare $c_name cursor for $p_name;

  chk_status("DECLARE(OPEN)", stmt);

  if (!sql_reopen(fd,argv)) {
    sql_close(fd);
    fd=-1;
  }
  chk_status("OPEN(OPEN)", stmt);

  ssql[fd].opened = 1;
  ssql[fd].insert_id=INSERT_ID;
  return fd;
}


int sql_error_code() {
  return SQLCODE;
}


bool sql_reopen(SQL_CURSOR fd, SQL_ROW &argv) {
  struct sqlda *udesc;
  $ char *c_name;

  if (!chk_cursor(fd, "sql_reopen"))
    return false;

  if (!sql_set_values(fd, argv))
    return false;

  udesc = ssql[fd].in;
  c_name = ssql[fd].c_name;

  if (udesc && udesc->sqld)
    $open $c_name using descriptor udesc;
  else
    $open $c_name;

  chk_status("OPEN","sql_reopen()")

  return (sqlca.sqlcode>=0);
}

/*
** If this is a non-cursory statement, sql_fetch() executes the statement.
** If this is a cursory statement, then it fetches a row of data.
*/
bool sql_fetch(SQL_CURSOR fd) {
  int         ret;
  struct sqlda *udesc;

  if (!chk_cursor(fd, "sql_fetch"))
    return false;

  if (!ssql[fd].is_select) {
    $char *p_name = ssql[fd].p_name;
    udesc = ssql[fd].in;
    if (udesc)
      $execute $p_name using descriptor udesc;
    else
      $execute $p_name;
    ret = sqlca.sqlcode;
    chk_status("EXECUTE(RUN)", ssql[fd].cmd);
  } else {
    $char *c_name = ssql[fd].c_name;
    udesc = ssql[fd].out;
    if (udesc && udesc->sqld)
      $fetch $c_name using descriptor udesc;
    else
      $fetch $c_name;
    ret = sqlca.sqlcode;
    chk_status("FETCH(OPEN)", ssql[fd].cmd);
  }
  ssql[fd].insert_id=INSERT_ID;
  return ((ret>=0) && (ret!=100));
}

/* Convert (blob) memory to hex encoding used in UNLOAD */
static char *convert_ascii(const char *str, int len) {
  int i;
  int j;
  char *hex = "0123456789ABCDEF";
  char *p;

  p = (char *)calloc(len*2+1, sizeof(char));
  if (!p) return p;

  /* BCD to ASCII legible ! */
  for (i = j = 0; i < len; i++) {
    p[j++] = hex[((str[i] & 0xF0) >> 4)];
    p[j++] = hex[(str[i] & 0x0F)];
  }
  p[j] = '\0';
  return p;
}

SQL_ROW sql_values(SQL_CURSOR fd, int *num, int dostrip) {
  static SQL_ROW argv;
  static int maxnum;
  struct sqlda *udesc;
  struct sqlvar_struct *col;
  int i;

  if (!chk_cursor(fd, "sql_values"))
    return NULL;
  if (!ssql[fd].is_select) {
    sprintf(errmsg,
            "Cannot get values for a non-select statement "
            "(fd %d) passed to sql_values", fd);
    debug(fprintf(stderr, "%s\n", errmsg));
    return NULL;
  }
  if (!ssql[fd].opened) {
    sprintf(errmsg,
            "Cannot get values for unopened statement "
            "(fd %d) passed to sql_values", fd);
    debug(fprintf(stderr, "%s\n", errmsg));
    return NULL;
  }
  udesc = ssql[fd].out;
  if (!udesc)
    return NULL;
  if (udesc->sqld > maxnum) {
    argv.argc(udesc->sqld);
    maxnum = udesc->sqld;
  }
  for (i = 0, col = udesc->sqlvar; i < udesc->sqld; i++, col++) {
    if (col->sqltype == CLOCATORTYPE) {
      loc_t *loc = (loc_t *)col->sqldata;
      /* loc->loc_user_env contains T or B */
      if (loc && loc->loc_buffer) {
        BLOBENV *benv;
        benv = (BLOBENV *)loc->loc_user_env;
        if (benv && benv->type == 'B') {
          char *oldptr;
          oldptr = benv->byte_ptr;
          benv->byte_ptr =
            convert_ascii(loc->loc_buffer, loc->loc_size);
          if (!benv->byte_ptr) argv[i] = "";
          else argv[i] = benv->byte_ptr;
          if (oldptr) free(oldptr);
        } else {
          argv[i] = loc->loc_buffer;
          if (loc->loc_size >= 0) argv[i][loc->loc_size] = '\0';
          if (dostrip == 1) strip(argv[i], loc->loc_size);
        }
      } else argv[i] = "";
    } else {
      argv[i] = const_cast<char *>((col->sqldata)?col->sqldata:"");
      if (dostrip == 1)
        strip(argv[i],strlen(argv[i]));
    }
  }
  if (num) *num = udesc->sqld;
  return argv;
}

SQL_ROW sql_sqlca(int *num) {
  static SQL_ROW argv((char *)0,8);
  static char sqlcode[15];
  static char sqlerrd[15*6];
  static char sqlwarn[25];
  static char sqlmsg[1024+1];
  static char isammsg[1024+1];
  int sqlmsg_len;
  int isammsg_len;

  sprintf(argv[0] = sqlcode, "%ld", sqlca.sqlcode);
  argv[1] = sqlca.sqlerrm;
  argv[2] = sqlca.sqlerrp;
  sprintf(argv[3] = sqlerrd, "%ld %ld %ld %ld %ld %ld",
          sqlca.sqlerrd[0], sqlca.sqlerrd[1], sqlca.sqlerrd[2],
          sqlca.sqlerrd[3], sqlca.sqlerrd[4], sqlca.sqlerrd[5]);
  sprintf(argv[4] = sqlwarn, "%c %c %c %c %c %c %c %c",
          sqlca.sqlwarn.sqlwarn0, sqlca.sqlwarn.sqlwarn1,
          sqlca.sqlwarn.sqlwarn2, sqlca.sqlwarn.sqlwarn3,
          sqlca.sqlwarn.sqlwarn4, sqlca.sqlwarn.sqlwarn5,
          sqlca.sqlwarn.sqlwarn6, sqlca.sqlwarn.sqlwarn7);

  if (sqlca.sqlcode == SQLNOTFOUND) {
    sprintf(sqlmsg, "Record not found.\n");
    sqlmsg_len = strlen(sqlmsg);
  } else if (sqlca.sqlcode == 0) {
    sqlmsg[0] = '\0';
    sqlmsg_len = 0;
  } else {
    rgetlmsg(sqlca.sqlcode, sqlmsg, sizeof(sqlmsg), &sqlmsg_len);
  }
  strip(sqlmsg,sqlmsg_len);
  if (sqlca.sqlerrd[1] == 0) {
    isammsg[0] = '\0';
    isammsg_len = 0;
  } else {
    rgetlmsg(sqlca.sqlerrd[1], isammsg, sizeof(isammsg), &isammsg_len);
  }
  strip(isammsg, isammsg_len);
  argv[5] = sqlmsg;
  argv[6] = isammsg;
  if (num)
    *num = 7;
  return argv;
}

int sql_sqld(SQL_CURSOR fd, int in) {
  struct sqlda *udesc;
  if (fd < 0 || fd >= sql::MAXCURSORS || !ssql[fd].cmd) {
    sprintf(errmsg, "Invalid fd %d passed to sql_sqld", fd);
    debug(fprintf(stderr, "%s\n", errmsg));
    return -2;
  }
  if (in == 1) udesc = ssql[fd].in;
  else if (in == 2) udesc = ssql[fd].exec;
  else udesc = ssql[fd].out;

  if (udesc) return udesc->sqld;
  return -1;
}

SQL_ROW sql_sqlcols(SQL_CURSOR fd, int in, char *arg, int *numvalues) {
  struct sqlda *udesc;
  struct sqlvar_struct *col;
  static SQL_ROW argv_names;
  static SQL_ROW argv_types;
  static SQL_ROW argv_charlen;
  static SQL_ROW argv_dblen;
  static SQL_ROW *argv;
  static int argv_len;
  int use_savedtypes = 0;
  int i;
  int colcmd = 0;

  if (!arg || *arg == '\0') {
    sprintf(errmsg,
            "Need argument: colnames,coltypes,colcharlen or coldblen");
    strcat(errmsg, "to sql_sqlcols");
    debug(fprintf(stderr, "%s\n", errmsg));
    return NULL;
  }
  if (strcmp(arg, "colnames") == 0) argv = &argv_names, colcmd = 1;
  else if (strcmp(arg, "coltypes") == 0) argv = &argv_types, colcmd = 2;
  else if (strcmp(arg, "colcharlen") == 0) argv = &argv_charlen, colcmd = 3;
  else if (strcmp(arg, "coldblen") == 0) argv = &argv_dblen, colcmd = 4;
  else {
    sprintf(errmsg,
            "Invalid argument %s to sql_sqlcols - need colnames,coltypes", arg);
    strcat(errmsg, ",colcharlen or coldblen");
    debug(fprintf(stderr, "%s\n", errmsg));
    return NULL;
  }

  if (fd < 0 || fd >= sql::MAXCURSORS || !ssql[fd].cmd) {
    sprintf(errmsg, "Invalid fd %d passed to sql_%s ", fd, arg);
    debug(fprintf(stderr, "%s\n", errmsg));
    return NULL;
  }

  if (in == 1) udesc = ssql[fd].in;
  else if (in == 2) udesc = ssql[fd].exec;
  else {
    udesc = ssql[fd].out;
    use_savedtypes = 1;
  }

  if (!udesc) {
    sprintf(errmsg, "Invalid IN # %d passed to sql_%s ", in, arg);
    debug(fprintf(stderr, "%s\n", errmsg));
    return NULL;
  }
  if (udesc->sqld <= 0) {
    sprintf(errmsg, "No %s available to sql_%s (sqld is %d)\n",
            arg, arg, udesc->sqld);
    debug(fprintf(stderr, "%s\n", errmsg));
    return NULL;
  }

  if (*argv == NULL || argv_len == 0 || argv_len < udesc->sqld) {
    if (*argv) {
      for (i = 0; i < argv_len; i++)
        if ((*argv)[i]) free((*argv)[i]);
      free(*argv);
    }
    if (argv_len < udesc->sqld) argv_len = udesc->sqld;
    *argv = (char *)calloc(argv_len, sizeof(char *));
    if (!(*argv)) {
      sprintf(errmsg, "Cannot malloc for %d in SQLDA errno %d",
              argv_len * sizeof(char *), errno);
      debug(fprintf(stderr, "%s\n", errmsg));
      return NULL;
    }
    for (i = 0; i < argv_len; i++)
      (*argv)[i] = (char *)calloc(30, sizeof(char));
  }

  for (i = 0, col = udesc->sqlvar; i < udesc->sqld; i++, col++) {
    switch (colcmd) {
      case 1:
        strcpy((*argv)[i] , col->sqlname);
        break;
      case 2:
        if (use_savedtypes == 1 && i < ssql[fd].arr_sqltype_len)
          sprintf((*argv)[i],"%s",rtypname(ssql[fd].arr_sqltype[i]));
        else
          sprintf((*argv)[i] , "%s", rtypname(col->sqltype));
        break;
      case 3:
        sprintf((*argv)[i] , "%ld",  col->sqllen);
        break;
      case 4:
        if (i < ssql[fd].arr_sqltype_len)
          sprintf((*argv)[i] , "%d",  ssql[fd].arr_sqllen[i]);
        else
          sprintf((*argv)[i], "%ld", col->sqllen);
        break;
      default:
        sprintf(errmsg, "Invalid argument to sql_%s", arg);
        return NULL;
    }
  }
  if (numvalues) *numvalues = udesc->sqld;
  return *argv;
}

SQL_ROW *sql_sqlda(SQL_CURSOR fd, int in, int num, int *numvalues) {
  struct sqlda *udesc;
  struct sqlvar_struct *col;
  static SQL_ROW argv((char *)0,11);
  static char sqltype[255];
  static char sqllen[15];
  static char sqlitype[255];
  static char sqlilen[15];
  int use_savedtypes = 0;
  static char sqltype_name[255];
  static char sqlitype_name[255];
  static char sqllen_num[15];

  if (!chk_cursor(fd, "sql_sqlda"))
    return NULL;

  if (in == 1) udesc = ssql[fd].in;
  else if (in == 2) udesc = ssql[fd].exec;
  else {
    udesc = ssql[fd].out;
    use_savedtypes = 1;
  }

  if (!udesc) {
    sprintf(errmsg, "Invalid IN # %d passed to sql_sqlda", in);
    debug(fprintf(stderr, "%s\n", errmsg));
    return NULL;
  }

  if (num < 0 || num >= udesc->sqld) {
    sprintf(errmsg, "Invalid num %d passed to sql_sqlda (>= %d)", num,
            udesc->sqld);
    debug(fprintf(stderr, "%s\n", errmsg));
    return NULL;
  }
  col = udesc->sqlvar + num;
  if (use_savedtypes == 1 && num < ssql[fd].arr_sqltype_len)
    sprintf(argv[0] = sqltype, "%d", ssql[fd].arr_sqltype[num]);
  else sprintf(argv[0] = sqltype, "%hd", col->sqltype);
  sprintf(argv[1] = sqllen, "%ld", col->sqllen);
  argv[2] = col->sqldata;
  /* Fix this up with col->sqltype&SQL etc. ! */
  if (!argv[2]) argv[2] = "";
  argv[3] = col->sqlname;
  if (!argv[3]) argv[3] = "";
  argv[4] = col->sqlformat;
  if (!argv[4]) argv[4] = "";
  sprintf(argv[5] = sqlitype, "%hd", col->sqlitype);
  sprintf(argv[6] = sqlilen, "%hd", col->sqlilen);

  if (use_savedtypes == 1 && num < ssql[fd].arr_sqltype_len)
    sprintf(argv[7] = sqltype_name, "%s",
            rtypname(ssql[fd].arr_sqltype[num]));
  else
    sprintf(argv[7] = sqltype_name, "%s", rtypname(col->sqltype));

  sprintf(argv[8] = sqlitype_name, "%s", rtypname(col->sqlitype));
  if (num < ssql[fd].arr_sqltype_len)
    sprintf(argv[9] = sqllen_num, "%d", ssql[fd].arr_sqllen[num]);
  else
    sprintf(argv[9] = sqllen_num, "%ld", col->sqllen);
  if (numvalues) *numvalues = 10;
  return &argv;
}

char *sql_command(SQL_CURSOR fd) {
  if (!chk_cursor(fd, "sql_command"))
    return NULL;
  return ssql[fd].cmd;
}

bool sql_close(SQL_CURSOR fd) {
  struct sqlda *udesc;
  struct sqlvar_struct *col;

  if (!ssql[fd].cmd)  return true;

  if (ssql[fd].is_select) {
    /* XXX - Error handling? */
    $char *c_name = ssql[fd].c_name;
    if (ssql[fd].opened)
      $close $c_name;
    $free $c_name;
  }

  {
    /* XXX - Error handling? */
    $char *p_name = ssql[fd].p_name;
    $free $p_name;
  }

  /* XXX - Valid to assert ssql[fd].cmd != 0 ? */
  if (ssql[fd].cmd && (ssql[fd].cmd != null_cmd))
    free(ssql[fd].cmd);

  ssql[fd].cmd = NULL;

  udesc = ssql[fd].in;
  if (udesc) {
    int         i;
    for (i = 0, col = udesc->sqlvar; i < udesc->sqld; i++, col++) {
      if (col->sqldata)
        free(col->sqldata), col->sqldata = NULL;
    }
    if (udesc->sqlvar)
      free(udesc->sqlvar), udesc->sqlvar = NULL;
    free(udesc), udesc = NULL;
  }
  ssql[fd].in = NULL;

  udesc = ssql[fd].out;
  if (udesc) {
    int         i;
    for (i = 0, col = udesc->sqlvar; i < udesc->sqld; i++, col++) {
      if (col->sqltype == CLOCATORTYPE) {
        loc_t      *loc = (loc_t *) col->sqldata;
        if (loc && loc->loc_buffer)
          free(loc->loc_buffer), loc->loc_buffer = NULL;
        if (loc && loc->loc_user_env) {
          BLOBENV    *benv = (BLOBENV *) loc->loc_user_env;
          if (benv->byte_ptr)
            free(benv->byte_ptr);
          benv->byte_ptr = NULL;
          free(loc->loc_user_env), loc->loc_user_env = NULL;
        }
      }
      if (col->sqldata)
        free(col->sqldata), col->sqldata = NULL;
      if (col->sqlind)
        free(col->sqlind), col->sqlind = NULL;
    }
    /* Should not free udesc->sqlvar and udesc since describe allocates
       space for it */
#ifdef  FIXBUG
    if (udesc->sqlvar)
      free(udesc->sqlvar), udesc->sqlvar = NULL;
    free(udesc);
    udesc = NULL;
#endif

  }

  udesc = ssql[fd].out;
  ssql[fd].out = NULL;
  /* If out and exec are the same then no need to the following */
  if (udesc != ssql[fd].exec && ssql[fd].exec) {
    int         i;
    udesc = ssql[fd].exec;
    for (i = 0, col = udesc->sqlvar; i < udesc->sqld; i++, col++) {
      if (col->sqltype == CLOCATORTYPE) {
        loc_t      *loc = (loc_t *) col->sqldata;
        if (loc && loc->loc_buffer)
          free(loc->loc_buffer), loc->loc_buffer = NULL;
        if (loc && loc->loc_user_env) {
          BLOBENV    *benv = (BLOBENV *) loc->loc_user_env;
          if (benv->byte_ptr)
            free(benv->byte_ptr);
          benv->byte_ptr = NULL;
          free(loc->loc_user_env), loc->loc_user_env = NULL;
        }
      }
      if (col->sqldata)
        free(col->sqldata), col->sqldata = NULL;
      if (col->sqlind)
        free(col->sqlind), col->sqlind = NULL;
    }
    /* Should not free udesc->sqlvar and udesc since describe allocates
       space for it */
#ifdef  FIXBUG
    if (udesc->sqlvar)
      free(udesc->sqlvar), udesc->sqlvar = NULL;
    free(udesc);
    udesc = NULL;
#endif

  }
  ssql[fd].exec = NULL;

  if (ssql[fd].arr_sqltype_len > 0) {
    free(ssql[fd].arr_sqltype);
    ssql[fd].arr_sqltype = NULL;
    free(ssql[fd].arr_sqllen);
    ssql[fd].arr_sqllen = NULL;
    ssql[fd].arr_sqltype_len = 0;
  }

  memset(&ssql[fd], 0, sizeof(SSQL));
  return (sqlca.sqlcode>=0);
}

bool sql_print(SQL_CURSOR fd) {
  struct sqlda *udesc;
  struct sqlvar_struct *col;

  if (fd < 0 || fd >= sql::MAXCURSORS) {
    sprintf(errmsg, "Invalid fd %d passed to sql_print", fd);
    debug(fprintf(stderr, "%s\n", errmsg));
    return false;
  }
  printf("Command string for fd %d: %s", fd,
         (ssql[fd].cmd?ssql[fd].cmd:"null"));
  if (!ssql[fd].opened) printf(" not ");
  printf(" opened, ");
  if (!ssql[fd].opened) printf(" not ");
  printf(" selected\n");
  if (ssql[fd].in) {
    int i;
    udesc = ssql[fd].in;
    printf("Input:");
    for (i = 0, col = udesc->sqlvar; i < udesc->sqld; i++, col++) {
      if (!(col->sqlname)) col->sqlname="null";
      printf("(%s,%ld,%s)\n",col->sqlname, col->sqllen, col->sqldata);
    }
  }
  if (ssql[fd].out) {
    int i;
    udesc = ssql[fd].out;
    printf("Output:\n");
    for (i = 0, col = udesc->sqlvar; i < udesc->sqld; i++, col++) {
      if (!(col->sqlname)) col->sqlname="null";
      printf("(%s,%ld,%s)\n",col->sqlname, col->sqllen, col->sqldata);
    }
  }
  return true;
}

bool sql_database(const char *dbname, const char *password, int exclusive) {
  $char *p;
  int len;
  int fd=alloc_stmt();

  if (!dbname || *dbname == '\0') {
    p = getenv("DATABASE");
    if (!p) p = "NODATABASE";
  } else
    p = const_cast<char *>(dbname);
  if (exclusive) {
    $database $p exclusive;
    chk_status("DATABASE EXCLUSIVE", p);
  } else {
    $database $p;
    chk_status("DATABASE", p);
  }
  if (database_name) {
    free(database_name);
    database_name = NULL;
  }
  len = strlen(p) + 1;
  database_name = (char *)calloc(len, 1);
  errmalloc(database_name, "DATABASE", len);
  strcpy(database_name, p);
  sql_close(fd);
  if (SQLCODE<0) {
    return false;
  }
  return true;
}

bool sql_connect(const char *dbname, const char *uid, const char *pw, const char *cname, int with_contrans) {
  $char *p;
  $char *username;
  $char *password;
  $char *conn_name;
  int fd=alloc_stmt();

  int len;
  if (!dbname || *dbname == '\0') {
    p = getenv("DATABASE");
    if (!p) p = "NODATABASE";
  } else p = const_cast<char *>(dbname);
  conn_name = const_cast<char *>(cname);
  username = const_cast<char *>(uid);
  password = const_cast<char *>(pw);
  if (cname && *cname != '\0') {
    if (with_contrans) {
      $connect to $p as $conn_name user $username using $password
      with concurrent transaction;
    } else {
      $connect to $p as $conn_name user $username using $password;
    }
  } else {
    if (with_contrans) {
      $connect to $p user $username using $password
      with concurrent transaction;
    } else {
      $connect to $p user $username using $password;
    }
  }
  chk_status("CONNECT", p);
  len = strlen(p) + 1;
  if (connection_name) {
    free(connection_name);
    connection_name = NULL;
  }
  connection_name = (char *)calloc(len, 1);
  errmalloc(connection_name, "CONNECT", len);
  strcpy(connection_name, p);
  sql_close(fd);
  return (SQLCODE>=0);
}

bool sql_disconnect(const char *arg) {
  $char *cname;

  cname = const_cast<char *>(arg);
  if (!arg) {
    sprintf(errmsg, "No argument passed to sql_disconnect");
    debug(fprintf(stderr, "%s\n", errmsg));
    return false;
  }
  if (strcmp(arg, "current") == 0) {
    $disconnect current;
  } else if (strcmp(arg, "default") == 0) {
  $disconnect default;
  } else if (strcmp(arg, "all") == 0) {
    $disconnect all;
  } else {
    $disconnect $cname;
  }
  int fd=alloc_stmt();
  chk_status("DISCONNECT", arg);
  sql_close(fd);
  return true;
}

bool sql_sigcmd(const char *arg) {
  $char *cname;

  cname = const_cast<char *>(arg);
  if (!arg) {
    sprintf(errmsg,
            "Argument must be one of sqlbreak, sqldetach, sqlexit, sqldone");
    debug(fprintf(stderr, "%s\n", errmsg));
    return false;
  }
  if (strcmp(arg, "sqlbreak") == 0) {
    return (!sqlbreak());
  } else if (strcmp(arg, "sqldetach") == 0) {
    return !sqldetach();
  } else if (strcmp(arg, "sqlexit") == 0) {
    return !sqlexit();
  } else if (strcmp(arg, "sqldone") == 0) {
    return !sqlexit();
  } else {
    sprintf(errmsg,
            "Argument must be one of sqlbreak, sqldetach, sqlexit, sqldone");
    debug(fprintf(stderr, "%s\n", errmsg));
    return false;
  }
}

bool sql_setconnection(const char *arg, int dormant) {
  $char *cname;

  cname =  const_cast<char *>(arg);
  if (!arg) {
    sprintf(errmsg, "No argument passed to sql_setconnection");
    debug(fprintf(stderr, "%s\n", errmsg));
    return false;
  }
  if (strcmp(arg, "current") == 0) {
    if (dormant)
      $set connection current dormant;
    else
      /* NOTE: COREDUMPs occur if dormant is removed and no connection is open !!!!*/
      /* HENCE DORMANT IS ALWAYS USED */
      $set connection current dormant;
  } else if (strcmp(arg, "default") == 0) {
    if (dormant)
    $set connection default dormant;
    else
    $set connection default;
  } else {
    if (dormant)
      $set connection $cname dormant;
    else
      $set connection $cname;
  }
  int fd=alloc_stmt();
  if (dormant) {
    chk_status("SET CONNECTION DORMANT", arg);
  } else {
    chk_status("SET CONNECTION", arg);
  }
  sql_close(fd);
  return true;
}

bool sql_finish(void) {
  if (database_name) {
    free(database_name);
    database_name = NULL;
  }
  $close database;
  int fd=0;
  chk_status("CLOSE", "DATABASE");
  return true;
}

bool sql_begin(void) {
  $begin work;
  int fd=alloc_stmt();
  chk_status("BEGIN", "WORK");
  sql_close(fd);
  return true;
}

int sql_commit(void) {
  $commit work;
  int fd=alloc_stmt();
  chk_status("COMMIT", "WORK");
  sql_close(fd);
  return true;
}

int sql_rollback(void) {
  $rollback work;
  int fd=alloc_stmt();
  chk_status("ROLLBACK", "WORK");
  sql_close(fd);
  return true;
}

char *sql_getdatabase(void) {
  return database_name;
}

/*
sql readblob table column "rowid = NUM" intofilename [append]
    if success (0) returns NULL or SIZE of blob written
    else returns ERROR (-1)
sql writeblob table column "rowid = NUM" [fromfilename | null] [size|-1]
    if success (0) returns SIZE of blob written
    else returns ERROR (-1)
*/
int sql_readblob(const char *table, const char *column, const char *rowidclause, const char *intofile,
                 int appendmode) {
  $loc_t sloc;
  $char stmt[sql::BUFSIZ+1];
  int fd=alloc_stmt();

  if (atoi(rowidclause) > 0)
    sprintf(stmt, "select %s from %s where rowid = %s",
            column, table, rowidclause);
  else
    sprintf(stmt, "select %s from %s where %s", column, table, rowidclause);

  sloc.loc_loctype = LOCFNAME;
  sloc.loc_fname  = const_cast<char *>(intofile);

  if (appendmode)
    sloc.loc_oflags = LOC_APPEND;
  else
    sloc.loc_oflags = LOC_WONLY;

  $prepare tcl_readblob from $stmt;
  chk_status("PREPARE(READBLOB)", stmt);

  $declare tcl_readblob_cur cursor for tcl_readblob;
chk_status("DECLARE(READBLOB)", stmt);

  $open tcl_readblob_cur;
  chk_status("OPEN(READBLOB)", stmt);

  $fetch tcl_readblob_cur into $sloc;
  chk_status("FETCH(READBLOB)", stmt);

  if (sqlca.sqlcode != 0) {
    sprintf(errmsg, "No record found.\n");
    debug(fprintf(stderr, "%s\n", errmsg));
    sql_close(fd);
    return -1;
  }

  /*
  After getting it
  sloc.loc_size # of bytes, sloc.loc_indicator = -1 for null,
  sloc.loc_status = 0 for success, -ve for error
  */
  if (sqlca.sqlcode == 0 && sloc.loc_status == 0) {
    if (sloc.loc_indicator == -1) {
      $close tcl_readblob_cur;
      sql_close(fd);
      return -2;
    }
    $close tcl_readblob_cur;
    sql_close(fd);
    return sloc.loc_size;
  }
  $close tcl_readblob_cur;
  sql_close(fd);
  
  return -1;
}

int sql_writeblob(const char *table, const char *column, const char *rowidclause, const char *fromfile,
                  int fromsize) {
  $loc_t uloc;
  $char stmt[sql::BUFSIZ+1];
  $char *colname;
  int fd=alloc_stmt();

  colname = const_cast<char *>(column);

  if (atoi(rowidclause) > 0)
    sprintf(stmt, "update %s set %s = ? where rowid = %s",
            table, column, rowidclause);
  else
    sprintf(stmt, "update %s set %s = ? where %s",
            table, column, rowidclause);

  /*
  sprintf(stmt, "select 1 from %s where %s FOR UPDATE OF %s",
      table, rowidclause, column);
  */

  uloc.loc_loctype = LOCFNAME;
  uloc.loc_fname = const_cast<char *>(fromfile);
  uloc.loc_oflags = LOC_RONLY;
  uloc.loc_size  = fromsize;
  uloc.loc_indicator = 0;

  if (fromfile)
    if (strcmp(fromfile,"null") == 0 || strcmp(fromfile, "NULL") == 0)
      uloc.loc_indicator = -1;

  $prepare tcl_writeblob from $stmt;
  chk_status("PREPARE(WRITEBLOB)", stmt);

  $execute tcl_writeblob using $uloc;
  chk_status("EXECUTE(WRITEBLOB)", stmt);

  /*
  $declare tcl_writeblob_cur cursor for tcl_writeblob;
  chk_status("DECLARE(WRITEBLOB)", stmt);

  $open tcl_writeblob_cur;
  chk_status("OPEN(WRITEBLOB)", stmt);

  $fetch tcl_writeblob_cur into :found;
  chk_status("FETCH(WRITEBLOB)", stmt);

  sprintf(stmt, "update %s set %s = ? where current of tcl_writeblob_cur",
      table, column);

  $prepare tcl_writeblob_new from $stmt;
  chk_status("PREPARE UPDATE(WRITEBLOB)", stmt);

  $execute tcl_writeblob_new using $uloc;
  chk_status("EXECUTE UPDATE(WRITEBLOB)", stmt);

  */

  /*
  After updating it
  uloc.loc_size # of bytes transferred
  uloc.loc_status = 0 for success, -ve for error
  */
  sql_close(fd);
  if (sqlca.sqlcode == 0 && uloc.loc_status == 0)
    return uloc.loc_size;

  return -1;
}

#ifdef USE_NAMESPACE
}
#endif

