// Copyright 2003 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "config.h"

#include <stdlib.h>
#include <string>
#include <errno.h>
#include <algorithm>
#include <vector>

#include "sqlrow.h"

SQL_ROW empty_row;

std::vector<char *> SQL_ROW::parse_type(char *&s) {
  char *open,*cls,*strt,*comma;
  char *q=strchr(s,'(');
  strt=q+1;
  int nopen=1;
  do {
    open=((strchr(q+1,'('))?strchr(q+1,'('):q+1+strlen(q+1));
    cls=((strchr(q+1,')'))?strchr(q+1,')'):q+1+strlen(q+1));
    if (cls<open) {
      nopen--;
      q=cls;
    } else {
      q=open;
      nopen++;
    }
  } while (nopen);
  if (*cls != 0) {
    *cls=0;
    cls++;
  }
  comma=((strchr(cls,','))?strchr(cls,','):cls+strlen(cls));
  *comma=0;
  std::vector<char *> rv(parse_row(strt));
  s=comma;
  return rv;
}

std::vector<char *> SQL_ROW::parse_list(char *&s) {
  char *cls,*strt,*open,*comma;
  char *q=strchr(s,'{');
  int nopen=1;
  strt=q+1;
  do {
    open=((strchr(q+1,'{'))?strchr(q+1,'{'):q+1+strlen(q+1));
    cls=((strchr(q+1,'}'))?strchr(q+1,'}'):q+1+strlen(q+1));
    if (cls<=open) {
      nopen--;
      q=cls;
    } else {
      q=open;
      nopen++;
    }
  } while (nopen);
  if (*cls != 0) {
    *cls=0;
    cls++;
  }
  comma=((strchr(cls,','))?strchr(cls,','):cls+strlen(cls));
  *comma=0;
  std::vector<char *> rv(parse_row(strt));
  s=comma;
  return rv;
}

char *SQL_ROW::parse_blob(char *&s) {
  char *rv=s;
  long long etag=strchr(s,'>')-s+1;
  if (etag>0) {
    char *p=strchr(s,'=');
    int tmp=strlen(s+etag);
    if (p && (p<(s+etag))) sscanf(p+1,"%d",&tmp);
    s+=(etag+tmp);
  } else {
    s+=(strlen(s));
  }
  return rv;
}
  

std::vector<char *> SQL_ROW::parse_row(char *&s) {
  char *p=s;
  char *q=p;
  char *comma, *dquote, *quote;
  char *s_end=s+strlen(s);
  std::vector<char *>::iterator i;
  std::vector<char *> rv;
  while (p && (p<s_end) && *p) {
    while (isspace(*p)) p++;
    while (*p==',') {
      *p=0;
      p++;
    }
    q=p;
    if (!strncmp(q,"LIST",4)) {
      std::vector<char *> tmprow(parse_list(q));
      for (i=tmprow.begin();i<tmprow.end();i++) {
        rv.push_back(*i);
      }
      p=q+1;
    } else if (!strncmp(q,"ROW",3)) {
      std::vector<char *> tmprow(parse_type(q));
      for (i=tmprow.begin();i<tmprow.end();i++) {
        rv.push_back(*i);
      }
      p=q+1;
    } else if (!strncmp(q,"<BYTE",5)) {
      char * blob=parse_blob(q);
      rv.push_back(blob);
      p=q+1;
    } else {
      comma=((strchr(p,','))?strchr(p,','):p+strlen(p));
      quote=((strchr(p,'\''))?strchr(p,'\''):p+strlen(p));
      dquote=((strchr(p,'\"'))?strchr(p,'\"'):p+strlen(p));
      p=std::min(std::min(comma,quote),dquote);
      switch (*p) {
        case ',':
	  rv.push_back(q);
          *p=0;
          p++;
          break;
	case '\'': 
	  if ((quote=strchr(p+1,'\''))) {
	    rv.push_back(p);
	    *(quote+1)=0;
	    p=quote+2;
	  } else {
	    rv.push_back(p);
            p=comma+1;
            *comma=0;
	  }
	  break;
	case '"':
	  if ((quote=strchr(p+1,'"'))) {
	    rv.push_back(p+1);
	    *quote=0;
	    p=quote+1;
	  } else {
	    rv.push_back(p);
	    p=comma+1;
            *comma=0;
	  }
	  break;
	default:
	  rv.push_back(q);
      }
    }
  }
  s=p;
  return rv;
}


SQL_ROW::SQL_ROW(const char *s,int nfields) {
  if (s) {
    args=new char [strlen(s)+1];
//    fprintf(stderr,"0x%p:SQL_ROW(const char *,int) allocated %d bytes at 0x%p\n",this,strlen(s)+1,args);
    strcpy(args,s);
  } else {
    args=0;
    return;
  }
  char *p=args;
  pval=parse_row(p);
}

SQL_ROW::SQL_ROW(const char **arr, int nfields) : pval(), args(0) {
  for (int i=0;i<nfields;i++) {
    char *p=new char[strlen(arr[i])+1];
//    fprintf(stderr,"0x%p:SQL_ROW(const char **,int) allocated %d bytes at 0x%p\n",this,strlen(arr[i])+1,args);
    strcpy(p,arr[i]);
    pval.push_back(p);
  }
}

SQL_ROW::SQL_ROW(const SQL_ROW &s,int start_col,int end_col) : args(0) {
  int i;
  char *p;
  if (end_col<start_col) end_col=s.argc()-1;
  for (i=start_col;i<(end_col+1);i++) {
    if (s[i]) {
      p=new char[strlen(s[i])+1];
//      fprintf(stderr,"0x%p:SQL_ROW(const SQL_ROW &,int, int) allocated %d bytes at 0x%p\n",this,strlen(s[i])+1,p);
      if (p) {
        strcpy(p,s[i]);
      }
    } else {
      p=0;
    }
    pval.push_back(p);
  }
}

SQL_ROW &SQL_ROW::operator =(const SQL_ROW &s) {
  if (this != &s) {
    args=0; 
    int i;
    char *p;
    for (i=0;i<s.argc();i++) {
      if (s[i]) {
        p=new char[strlen(s[i])+1];
//        fprintf(stderr,"0x%p:operator = allocated %d bytes at 0x%p\n",this,strlen(s[i])+1,p);
        if (p) {
          strcpy(p,s[i]);
        }
      } else {
        p=0;
      }
      pval.push_back(p);
    }
  }
  return *this;
}

SQL_ROW::~SQL_ROW() {
  if (args) {
    delete args;
//    fprintf(stderr,"0x%p: freed %p\n",this,args);
  } else {
    int i;
    for (i=0;i<argc();i++) {
      delete pval[i];
//      fprintf(stderr,"0x%p: freed %p\n",this,pval[i]);
    }
  }
}

