// Copyright 2003 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef _SQLROW_H
#define _SQLROW_H

#undef swap
#include <cstdio>
#include <string>
#include <cstdlib>
#include <vector>
#include <algorithm>
#include "xml_util.h"

typedef int SQL_CURSOR;

class SQL_ROW;
extern SQL_ROW empty_row;

class SQL_ROW {
  public:
    SQL_ROW(const char *s=NULL,int i=0);
    explicit SQL_ROW(const char **arr, int nfields);
    SQL_ROW(const SQL_ROW &r, int start_col=0, int end_col=-1);
    ~SQL_ROW();
    SQL_ROW &operator=(const SQL_ROW &r);
    std::vector<char *> pval;
    char *args;
    int argc(int i);
    int argc() const;
    operator char **();
    SQL_ROW operator +(int i) const;
    char *&operator [](int i);
    const char *operator [](int i) const;
//    char *operator *() const;
  private:
    std::vector<char *> parse_type(char *&s);
    std::vector<char *> parse_list(char *&s);
    std::vector<char *> parse_row(char *&s);
    char *parse_blob(char *&s);
};

extern SQL_ROW empty_row;

inline SQL_ROW::operator char **() {return static_cast<char **>(&(pval[0]));}

inline char *&SQL_ROW::operator [](int i) {return (*(pval.begin()+i));}

inline const char *SQL_ROW::operator [](int i) const {
  if (((unsigned)i)<pval.size()) {
    return (*(pval.begin()+i));
  } else {
    return "";
  }
}

//inline char *SQL_ROW::operator *() const {return (pval[0]);}

inline int SQL_ROW::argc(int i) {pval.resize(i); return (int)pval.size();}

inline int SQL_ROW::argc() const {return (int)pval.size();}

inline SQL_ROW SQL_ROW::operator +(int i) const { return SQL_ROW(*this,i);  }

#endif
